local ADDON, data = ...

local BOX_W = 245
local BOX_LEFT_W = 104
local BOX_H = 21
local MAX_CHILDREN = 7

tttt = {
    [1] = true,
    [2] = 'mooo',
    [3] = 3,
    ['test'] = 123,
    ['hello'] = 321,
    ['moo'] = 123123,
    [4] = {
        [1] = 'abcdefg',
        ['abcdefg'] = 1,
    },
	['4'] = 'booyah',
	[{[1] = 1}] = function () end,
	[function () end] = {
        [1] = 'abcdefg',
        ['abcdefg'] = 1,
	},
	[true] = true,
	

}
----------------------------------------------------------------------------

local function SetPrettyBorder (frm)
    frm:SetBackdrop (GameTooltip:GetBackdrop())
    frm:SetBackdropColor (GameTooltip:GetBackdropColor())
    frm:SetBackdropBorderColor (GameTooltip:GetBackdropBorderColor())
end
----------------------------------------------------------------------------

local function RemovePrettyBorder (frm)
    frm:SetBackdrop (nil)
    frm:SetBackdropColor (nil)
    frm:SetBackdropBorderColor (nil)
end
----------------------------------------------------------------------------

do
    local function GetTableValue (tblName, keys)
        local tmp = _G[tblName]
        if type(tmp)~='table' then
            return nil
        end
        
        for i,key in ipairs(keys) do
            tmp = tmp[key]
        end
        
        return tmp
    end
    ----------------------------------------------------------------------------
    
    local function ValueToString (val)
        if type(val)=='table' then
            -- Is this table a frame?
            if val.GetObjectType then
                return string.format ('(%s)', val:GetObjectType())
            else
                local keys = 0
                local subtables = 0
                for k,v in pairs(val) do
                    keys = keys + 1
                    if type(v)=='table' then
                        subtables = subtables + 1
                    end
                end
                --return string.format ("%d keys\n%d sub-tables", keys, subtables)
                return string.format ("(table - %d keys)", keys)
            end
        elseif type(val)=='function' then
            return '(function)'
        elseif type(val)=='userdata' then
            return '(userdata)'
        else
            return tostring(val)
        end
    end
    ----------------------------------------------------------------------------
    
    local function KeyToString (val)
        if type(val)=='table' then
            return '(table)'
        elseif type(val)=='function' then
            return '(function)'
        elseif type(val)=='userdata' then
            return '(userdata)'
        elseif type(val)=='string' then
            return "'"..val.."'"
        else
            return tostring(val)
        end
    end
    ----------------------------------------------------------------------------

    local function CreateTextbox (parent)
        local f = parent:CreateFontString(nil, "HIGH", "GameFontNormal")
        f:SetTextColor (1.0,0.77,0.27,1.0)
        f:SetFont ("Fonts\\FRIZQT__.TTF", 10)
        
        return f
    end
    ----------------------------------------------------------------------------
    
    function CreateWatchBox (idx)
        local f = CreateFrame ("Frame", 'frmWatch'..tostring(idx), UIParent)
        f.idx = idx
        f:SetSize (BOX_W, BOX_H)
        --[[f:SetBackdrop ({
            bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
            edgeSize = 14,
        })]]

        -- Textboxes
        f.txtExpand = CreateTextbox(f)
        --f.txtExpand:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -4, 0)
        f.txtExpand:SetPoint ("BOTTOMRIGHT", f, "BOTTOMRIGHT", -4, 2)
        f.txtExpand:SetSize (10, BOX_H)
        f.txtExpand:SetFont ("Fonts\\FRIZQT__.TTF", 16)
        f.txtExpand:SetText ('+')
        f.txtExpand:SetJustifyH ("RIGHT")
        
        f.txtLeft = CreateTextbox(f)
        --f.txtLeft:SetPoint ("TOPLEFT", f, "TOPLEFT", 6, 0)
        f.txtLeft:SetPoint ("BOTTOMLEFT", f, "BOTTOMLEFT", 6, 3)
        f.txtLeft:SetSize (BOX_LEFT_W, BOX_H)
        f.txtLeft:SetJustifyH ("LEFT")
        
        f.txtRight = CreateTextbox(f)
        f.txtRight:SetPoint ("TOPLEFT", f.txtLeft, "TOPRIGHT")
        f.txtRight:SetPoint ("BOTTOMRIGHT", f.txtExpand, "BOTTOMLEFT", 6, 1)
        f.txtRight:SetJustifyH ("LEFT")
        
        -- Create up/down buttons, keep them hidden till they're needed
        f.btnUp = CreateFrame ("Button", nil, f)
        f.btnUp:SetSize (BOX_W - 10, 16)
        f.btnUp:SetNormalTexture ("Interface\\Addons\\"..ADDON.."\\up")
        f.btnUp:SetHighlightTexture ("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")
        f.btnUp:Hide()
        f.btnUp:SetScript ("OnClick", function (self)
            self:GetParent():ScrollUp()
        end)
        f.btnDown = CreateFrame ("Button", nil, f)
        f.btnDown:SetSize (BOX_W - 10, 16)
        f.btnDown:SetNormalTexture ("Interface\\Addons\\"..ADDON.."\\down")
        f.btnDown:SetHighlightTexture ("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")
        f.btnDown:Hide()
        f.btnDown:SetScript ("OnClick", function (self)
            self:GetParent():ScrollDown()
        end)
        
        -- The border around the child frames
        f.childBorder = CreateFrame ("Frame", nil, parent)
        SetPrettyBorder (f.childBorder)
        
        ------------------------------
        
        function f:GetRootFrame ()
            if self:GetParent()==UIParent then
                return self
            else
                return self:GetParent():GetRootFrame()
            end
        end        
        ------------------------------
        
        function f:CreateTag (title)
			-- First, some size adjustments for frames that have a tag...
            self:SetWidth (self:GetWidth() + 4)
			self:SetHeight (self:GetHeight() + 4+15)
			self.childOffsetX = -3
			self.childOffsetY = -4-15
			
			-- Now, that tag...
            local f = CreateFrame ("Frame", nil, self)
            f:SetPoint ("TOP", self, "TOP")
            f:SetSize (BOX_W, 16)
			--SetPrettyBorder (f)
            
            -- X button
            f.btnX = CreateFrame ("EditBox", nil, f)
            f.btnX:SetPoint ("TOPRIGHT", f, "TOPRIGHT")
            f.btnX:SetPoint ("BOTTOMLEFT", f, "BOTTOMRIGHT", -13, 1)
            f.btnX:SetTextColor (1.0,0.77,0.27,1.0)
            f.btnX:SetTextInsets (0, 0, 2, 2)
            f.btnX:SetFont ("Fonts\\FRIZQT__.TTF", 10)
            f.btnX:SetAutoFocus (false)
            f.btnX:Disable ()
            f.btnX:SetText ('x')
            f.btnX:SetScript ("OnMouseDown", function (self, button)
                self:GetParent():GetParent():Free()
            end)            
            
            -- Label
            f.lbl = f:CreateFontString(nil, "HIGH", "GameFontNormal")
            f.lbl:SetPoint ("TOPLEFT", f, "TOPLEFT", 8, -6)
            f.lbl:SetPoint ("BOTTOMRIGHT", f.btnX, "BOTTOMLEFT")
            f.lbl:SetTextColor (1.0,0.77,0.27,1.0)
            f.lbl:SetFont ("Fonts\\FRIZQT__.TTF", 11)
            f.lbl:SetText (title--[[self.varName]])
			--f.lbl:SetJustifyH ("LEFT")

            -- Make movable... maybe
            if true --[[isMovable]] then
                self:SetMovable (true)
                --f:RegisterForDrag ("LeftButton")
                self:SetClampedToScreen (true)
                self:EnableMouse (true)
                self:EnableMouseWheel (true)
                f:SetScript("OnMouseDown", function(self, button)
                    if button=='LeftButton' then
                        self:GetParent():StartMoving()
                    end
                end)
                f:SetScript("OnMouseUp", function(self, button)
                    if button=='LeftButton' then
                        self:GetParent():StopMovingOrSizing()
                    end
                end)
            end
            
            self.tag = f
            return f
        end
        
        ------------------------------
        
        function f:SetTableView (isTable)
            self.isTable = isTable
            
            if isTable then
                -- List of child values
                self.childKeys = {}
				
				-- Sort the keys
				local keysByType = {}
				for k,v in pairs(self.value) do
                    keysByType[type(k)] = keysByType[type(k)] or {}
					table.insert (keysByType[type(k)], k)
                end
				
				local orderedTypes = {'table', 'function', 'number', 'string', 'boolean'}
				for _,_type in ipairs(orderedTypes) do
					if keysByType[_type] then
						if _type=='string' or _type=='number' then
							table.sort (keysByType[_type])
						end
						for k,v in pairs(keysByType[_type]) do
							table.insert (self.childKeys, v)
						end
					end
					keysByType[_type] = nil
				end
				for i,tbl in ipairs(keysByType) do
					for k,v in pairs(keysByType[_type]) do
						table.insert (self.childKeys, v)
					end
				end
				
				--
                
                self.txtExpand:SetWidth(30)
                self.isExpanded = false
                self:Collapse()
            else
                self.txtExpand:SetWidth(1)
            end
        end
        ------------------------------
        
        function f:LoadSimpleVar (varName)
            -- varName: the name of a global var
            self.varName = varName
            local v = _G[varName]
            
            self.value = v
            self.txtRight:SetText (ValueToString(v))
            self.txtLeft:SetText ('('..type(v)..')')
            
            self:CreateTag (varName)
            self:SetTableView (type(v)=='table')
        end
        ------------------------------
        
        function f:LoadVar (varName)
            local keys = {}
            --for v in string.gmatch (varName, "%[(.-)%]") do
            for v in string.gmatch (varName, "[^%.%[%]]+") do
               tinsert (keys, v)
            end
            table.remove (keys, 1)
            if (#keys==0) then
                self:LoadSimpleVar (varName)
                return
            end
            
            self.varName = string.match (varName, "^[^%[%]%.]+")
            self.keys = keys
            
            local v = GetTableValue (self.varName, self.keys)
            self.value = v
            self.txtLeft:SetText ('('..type(v)..')')
            self.txtRight:SetText (ValueToString(v))
            
            self:CreateTag (varName)
            self:SetTableView (type(v)=='table')
        end
        ------------------------------
        
        function f:LoadTableElement (varName, key)
            self.varName = varName
            self.keys = {}
            for i,v in ipairs(self:GetParent().keys) do
                self.keys[i] = self:GetParent().keys[i]
            end
            table.insert (self.keys, key)
            
            self.txtLeft:SetText (KeyToString(key))
            local v = GetTableValue (varName, self.keys)
            self.value = v
            self.txtRight:SetText (ValueToString(v))
            
            self:SetTableView (type(v)=='table')
        end
        ------------------------------
        
        function f:ShowChildrenFrom (first)
            self:CollapseChildren()
            
            -- Bugger counting... this loop can just go till it runs out of frames
            local cidx = 1  -- child index
            for i=first,first+#self.childKeys-1 do
                if (not self.childFrames[cidx]) or (i>#self.childKeys) then break end
                self.childFrames[cidx]:LoadTableElement (self.varName, self.childKeys[i])
                cidx = cidx + 1
            end
            
            self.firstShownChild = first

            self.btnUp:SetShown (first>1)
            self.btnDown:SetShown (first+MAX_CHILDREN-1 < #self.childKeys)
                
        end        
        ------------------------------

        function f:ScrollUp ()
            if self.btnUp:IsShown() and #self.childKeys>MAX_CHILDREN then
                self.firstShownChild = self.firstShownChild - 1
                f:ShowChildrenFrom (self.firstShownChild)
            end
        end
        
        function f:ScrollDown ()
            if self.btnDown:IsShown() and #self.childKeys>MAX_CHILDREN then
                self.firstShownChild = self.firstShownChild + 1
                f:ShowChildrenFrom (self.firstShownChild)
            end
        end
        ------------------------------
        
        function f:Expand ()
            -- Collapse all siblings
            if #self.keys > 0  and  self:GetParent()~=UIParent then
                local p = self:GetParent()
                p:CollapseChildren()
                --p:Expand()
            end
            
            -- Wait, does this frame even have children?
            if #self.childKeys == 0 then
                GameTooltip:SetOwner (self, "ANCHOR_CURSOR")
                GameTooltip:AddLine ('This table is empty!', 1.0, 1.0, 0.8)
                GameTooltip:Show ()
                return
            end            
            
            -- Show children
            for i=1, min(MAX_CHILDREN, #self.childKeys) do
                local tmp = poolHandler.get()
                tmp:SetParent (self)
                tmp:SetPoint ("TOPLEFT", self, "TOPRIGHT", (self.childOffsetX or 0) + 4, -(i-1) * BOX_H + (self.childOffsetY or 0))
                
                tmp:Show()
                table.insert (self.childFrames, tmp)                                
            end
            
            self.childBorder:SetPoint ("TOPLEFT", self.childFrames[1], "TOPLEFT", -3, 4)
            self.childBorder:SetPoint ("BOTTOMRIGHT", self.childFrames[#self.childFrames], "BOTTOMRIGHT", 3, -2)
            self.childBorder:Show ()
            
            self:ShowChildrenFrom (self.firstShownChild or 1)
            
            -- Place up/down buttons
            if #self.childKeys > MAX_CHILDREN then
                self.btnUp:SetPoint ("BOTTOM", self.childBorder, "TOP", 0, -2)
                self.btnDown:SetPoint ("TOP", self.childBorder, "BOTTOM", 0, 2)
            end
            
            self.txtExpand:SetText ('-')
            self.isExpanded = true
        end
        ------------------------------
        
        function f:Collapse ()
            if not self.isTable then return end
        
            self.txtExpand:SetText ('+')
            self.isExpanded = false
            
            if #self.childKeys>MAX_CHILDREN then
                self.btnUp:Hide()
                self.btnDown:Hide()
            end
            
            for i,v in ipairs(self.childFrames) do
                --v:Collapse()
                v:Free()
                self.childFrames[i] = nil
            end
            
            self.childBorder:ClearAllPoints ()
            self.childBorder:Hide ()
        end
        ------------------------------
        
        function f:ShowTooltip ()
            GameTooltip:SetOwner (self, "ANCHOR_CURSOR")
            GameTooltip:AddLine (self.txtLeft:GetText(), 1,1,0)
            GameTooltip:AddLine (self.txtRight:GetText(), 1,1,1)
            GameTooltip:Show ()
        end
        ------------------------------

        function f:CollapseChildren ()
            for i,v in ipairs(self.childFrames) do
                self.childFrames[i]:Collapse()
            end
        end
        ------------------------------
        -- Events
        
        f:SetScript ("OnMouseDown", function (self, button,...)
            if self.isTable then
                if self.isExpanded then
                    self:Collapse ()
                else
                    self:Expand ()
                end
            end
        end)
        ------------------------------
        
        f:SetScript("OnMouseWheel", function (self, delta)
            if IsControlKeyDown() then
                self:GetRootFrame():SetScale (self:GetRootFrame():GetScale() + delta*0.04)
            else
                if delta>0 then 
                    if self:GetParent().ScrollUp then self:GetParent():ScrollUp() end
                else
                    if self:GetParent().ScrollDown then self:GetParent():ScrollDown() end
                end
            end
        end)        
        ------------------------------

        f:SetScript ("OnEnter", function (self)
            if (f.txtLeft:GetStringWidth() > f.txtLeft:GetWidth()  or  f.txtRight:GetStringWidth() > f.txtRight:GetWidth()) then
                self:ShowTooltip ()
            end
        end)

        f:SetScript ("OnLeave", function (self)
            GameTooltip:Hide ()
        end)
        ------------------------------
        
        function f:Free ()
            -- Free children
            self:Collapse()
            
            -- Get rid of the extra controls
            if self.tag then
                self.tag:Hide()
                self.tag = nil
            end
            self.btnUp:Hide()
            self.btnDown:Hide()
            
            self.varName = nil
            self.value = nil            
            
            self:SetParent (UIParent)
            self:Hide()
            
            poolHandler.free (self.idx)
        end
        
        ------------------------------
        
        function f:Init ()
            self.varName = nil
            self.keys = {}
            -- A box showing g_tbl['druid']['balance'] will have f.keys=={'druid','balance'} and f.varName='g_tbl'
            self.free = false
            self.childFrames = {}
            self.childKeys = {}
            self.isExpanded = nil
            self.firstShownChild = 1
			self:SetSize (BOX_W, BOX_H)
            self:SetBackdrop (nil)
			self:SetBackdropColor (nil)
			self:SetBackdropBorderColor (nil)
			self.childOffsetX = 0
			self.childOffsetY = 0
        end
        ------------------------------
        
        function f:InitWatchFrame ()
			self:SetTableView (false)
        end
        
        ------------------------------
        
        function f:LoadValue (v)
            self.value = v
            self.txtRight:SetText (tostring(v))
            self.txtLeft:SetText ('('..type(v)..')')
        end
        
        ------------------------------
        
        function f:AdjustWidth (w)
            local delta = self:GetWidth() - w
            self:SetWidth (w)
            self.tag:SetWidth (self.tag:GetWidth() - delta)
        end
        
        ------------------------------
        
        f:Hide()
        f:Init()
        f.free = true
        
        return f
    end
end

------------------------------------------------------------------

do
    local pool = {}
    local countFree = 0
    
    local function GetSpecificBox (n)
        if not pool[n].free then
            print ('Error: box '..tostring(n)..' is not free')
            return nil
        end
        
        pool[n]:Init()
        pool[n]:Show()
        countFree = countFree - 1
        return pool[n]
    end
    
    local function GetBox ()
        if countFree > 0 then
            for i=1,#pool do
                if pool[i].free then
                    return GetSpecificBox(i)
                end
            end
        else
            local tmp = CreateWatchBox(#pool+1)
            table.insert (pool, tmp)
            countFree = countFree + 1
            return GetSpecificBox(#pool)
        end
        
        print ('WatchIt Weird Error Code 10001')
    end
    
    local function FreeBox (idx)
        if pool[idx].free then
            --local errcode = 10000+idx
            --print ('WatchIt Weird Error Code '..errcode)
            return 
        end
        
        countFree = countFree + 1
        pool[idx].free = true
        pool[idx]:Hide()
    end
    
    local function ResetPool (idx)
        for i=1,#pool do
            pool[i]:Hide()
            pool[i] = nil
        end
        countFree = 0
    end
    
    poolHandler = {
        get = GetBox,
        free = FreeBox,
        reset = ResetPool,
        report = function ()
            print ('Pool size: '..#pool)
            print ('Free boxes: '..countFree)
        end
    }

end

------------------------------------------------------------------

function WatchVar (varName)
    local f = poolHandler.get()
    f:SetPoint ("TOPLEFT", UIParent, "TOPLEFT", 100, -150)
    f:LoadVar (varName)
    SetPrettyBorder (f)

    return f
end

------------------------------------------------------------------

function CreateWatchFrame (title)
    local f = poolHandler.get()
    f:SetPoint ("TOPLEFT", UIParent, "TOPLEFT", 100, -150)
    f.varName = title
    --f:CreateTag (true)
    f:InitWatchFrame ()
    
    return f
end

------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Slash commands

SLASH_WATCHIT1 = "/watch"
SlashCmdList["WATCHIT"] = function (txt)
    WatchVar (txt)
end
